"""
Class for calling IBD from Posterior Data. Saves results as tab seperated .tsv
created by Pandas and Numpy
Contains Sub-Classes, as well as factory Method.
Pls always change parameters with set_params method!
@ Author: Harald Ringbauer, 2020
"""

import numpy as np
import pandas as pd
import os as os

class PostProcessing(object):
    """Class that can do PostProcessing of HAPSBURG output.
    (for one individual). Sometimes post-processing is done outside that,
    Has Methods to save the output. Saves using standard hapROH format"""
    folder = ""          # The Folder to operate in.
    cutoff_post = 0.99    # Cutoff Probability for ROH State
    min_cm = 4      # Cutoff of block length [in cM]
    max_gap = 0.01  # The Maximum Gap Length to be Merged [in Morgan]
    save = 0        # What to save. 0: Nothing 1: Save post-processed IBD. 2: Save 0-posterior. 3: Save full posterior
    output = True   # Whether to plot output
    ch = 3            # Chromosome to analyze
    iid = "iid0"

    def __init__(self, folder=""):
        """Initialize Class.
        Load: Whether to immediately Load the Posterior Data"""
        self.folder=folder 
        pass

    def set_params(self, **kwargs):
        """Set the Parameters.
        Takes keyworded arguments"""
        for key, value in kwargs.items():
            setattr(self, key, value)
            
    def roh_posterior(self, posterior0):
        """Calculate the IBD posterior.
        Input: Posterior 0 [l]
        Output: 1 - Posterior 0 [l]"""
        #roh_post = 1 - np.exp(posterior0)  # Go to non-logspace probability
        roh_post = 1 - posterior0
        return roh_post
    
    def ibd_stat_to_block(self, ibd):
        """Convert IBD status per marker
        into list of ibd.
        Input: IBD stats [l] boolean.
        Return start and end indexes of IBD blocks"""
        x1 = np.hstack([[False], ibd, [False]]).astype("int")  # padding
        d = np.diff(x1)
        starts = np.where(d == 1)[0]
        ends = np.where(d == -1)[0]
        return starts, ends
    
    def create_df(self, starts, ends, starts_map, ends_map, 
              l, l_map, ch, min_cm, iid1, iid2=""):
        """Create and returndthe hapBLOCK/hapROH dataframe."""

        full_df = pd.DataFrame({'Start': starts, 'End': ends,
                                'StartM': starts_map, 'EndM': ends_map, 'length': l,
                                'lengthM': l_map, "ch": ch, 'iid1': iid1, "iid2": iid2})
        df = full_df[full_df["lengthM"] > min_cm/100.0]  # Cut out long blocks
        return df
    
    def merge_called_blocks(self, df, max_gap=0):
        """Merge Blocks in Dataframe df and return merged Dataframe"""
        if len(df) == 0:
            return df  # In case of empty dataframe don't do anything

        if max_gap == 0:
            max_gap = self.max_gap

        df_n = df.drop(df.index)  # Create New Data frame with all raws removed
        row_c = df.iloc[0, :].copy()

        # Iterate over all rows, update blocks if gaps small enough
        for index, row in df.iterrows():
            if row["StartM"] - row_c["EndM"] < max_gap:
                row_c["End"] = row["End"]
                row_c["EndM"] = row["EndM"]
                row_c["length"] = row_c["End"] - row_c["Start"]
                row_c["lengthM"] = row_c["EndM"] - row_c["StartM"]

            else:  # Save and go to next row
                df_n.loc[len(df_n)] = row_c  # Append a row to new df
                row_c = row.copy()

        df_n.loc[len(df_n)] = row_c   # Append the last row

        if self.output:
            print(f"Merged n={len(df) - len(df_n)} gaps < {max_gap} M")
        return df_n
    
    def save_output(self, df, r_map=[], post=[], save_folder=""):
        """Save hapBLOCK output in standardized format."""
        if len(save_folder)==0:
            save_folder = self.folder
            
        if not os.path.exists(save_folder):
            os.makedirs(save_folder)
            if self.output:
                print(f"Created {save_folder}.")
            
        path_ibd = os.path.join(save_folder, "ibd.tsv")
        self.save_ibd_df(df_ibd=df, save_path = path_ibd)

        if len(r_map)>0:
            path_map = os.path.join(save_folder, "map.tsv")
            np.savetxt(path_map, r_map, delimiter="\t")
        if len(post)>0:
            path_posterior = os.path.join(save_folder, "posterior.tsv")
            np.savetxt(path_posterior, post, delimiter="\t")
        if self.output:
            print(f"Successfully saved all output to {save_folder}")
            
    def save_ibd_df(self, df_ibd, save_path):
        """Save IBD Dataframe to path save_path"""
        df_ibd.to_csv(save_path, sep="\t", index=False)
        if self.output:
            print(f"Saved IBD output to: {save_path}")
        
    def call_roh(self, r_map, post0, iid1="", iid2=""):
        """Call ROH of Homozygosity from Posterior Data
        bigger than cutoff.
        post0: posterior in format [5,l], log space"""
        ibd_post = self.roh_posterior(post0[0,:])
        ibd = ibd_post > self.cutoff_post
        
        if len(iid1)==0:
            iid1=self.iid

        if self.output:
            frac_ibd = np.mean(ibd)
            print(f"Fraction Markers above IBD cutoff: {frac_ibd:.4f}")

        # Identify Stretches by difference (up and down)
        starts, ends = self.ibd_stat_to_block(ibd)
        l = ends - starts
        ends_map = r_map[ends - 1]  # -1 to stay within bounds
        starts_map = r_map[starts]
        l_map = ends_map - starts_map

        # Create hapROH Dataframe
        df = self.create_df(starts, ends, starts_map, ends_map, 
                            l, l_map, self.ch, min_cm=self.min_cm,
                            iid1=iid1, iid2=iid2)

        # Merge Blocks in Postprocessing Step
        if self.max_gap>0:
            df = self.merge_called_blocks(df)

        if self.output:
            print(f"Called n={len(df)} IBD Blocks > {self.min_cm} cM")
            l = np.max(df["lengthM"])
            print(f"Longest Block: {l *100:.2f} cM")

        if self.save==1:
            self.save_output(df)
        elif self.save==2:
            self.save_output(df, r_map=r_map, post=post0[0,:])
        elif self.save==3:
            self.save_output(df, r_map=r_map, post=post0)
        return df, r_map, post0
    
class NoPostProcessing(PostProcessing):
    """Class that does no Postprocessing"""
    
    def call_roh(*prms, **kwd):
        """Does nothing"""
        pass

class IBD2Postprocessing(PostProcessing):
    min_cm2 = 2.0
    def create_df(self, starts, ends, starts_map, ends_map, 
              l, l_map, ch, min_cm, iid1, iid2, segment_type="IBD1"):
        """Create and returndthe hapBLOCK/hapROH dataframe."""
        segment_types = [segment_type]*len(starts)
        full_df = pd.DataFrame({'iid1': iid1, "iid2": iid2, "ch": ch,
                                'Start': starts, 'End': ends, 'length': l,
                                'StartM': starts_map, 'EndM': ends_map, 'lengthM': l_map,
                                'segment_type': segment_types})
        df = full_df[full_df["lengthM"] > min_cm/100.0]  # Cut out long blocks
        return df

    def get_posterior(self, post0):
        ibd1 = np.sum(post0[1:, :], axis=0)
        ibd2 = np.sum(post0[5:, :], axis=0)
        return ibd1, ibd2


    def call_roh(self, r_map, post0, iid1="", iid2=""):
        """Call ROH of Homozygosity from Posterior Data
        bigger than cutoff.
        post0: posterior in format [7,l], log space"""
        assert(post0.shape[0] == 7)
        ibd1_post, ibd2_post = self.get_posterior(post0)
        ibd1 = ibd1_post > self.cutoff_post
        ibd2 = ibd2_post > self.cutoff_post
        
        if len(iid1)==0:
            iid1=self.iid

        if self.output:
            frac_ibd1 = np.mean(ibd1)
            print(f"Fraction Markers above IBD1 cutoff: {frac_ibd1:.4f}")
            frac_ibd2 = np.mean(ibd2)
            print(f'Fraction Markers above IBD2 cutoff: {frac_ibd2:.4f}')

    ############################ Writing IBD1 blocks to pandas dataframe ############################
        # Identify Stretches by difference (up and down)
        starts_ibd1, ends_ibd1 = self.ibd_stat_to_block(ibd1)
        l_ibd1 = ends_ibd1 - starts_ibd1
        ends_map_ibd1 = r_map[ends_ibd1 - 1]  # -1 to stay within bounds
        starts_map_ibd1 = r_map[starts_ibd1]
        l_map_ibd1 = ends_map_ibd1 - starts_map_ibd1

        # Create hapROH Dataframe
        df1 = self.create_df(starts_ibd1, ends_ibd1, starts_map_ibd1, ends_map_ibd1, 
                            l_ibd1, l_map_ibd1, self.ch, self.min_cm,
                            iid1, iid2, segment_type='IBD1')
        if self.output:
            print(f"Called n={len(df1)} IBD1 Blocks > {self.min_cm} cM")
            l = np.max(df1["lengthM"])
            print(f"Longest Block: {l *100:.2f} cM")
        # Merge Blocks in Postprocessing Step
        if self.max_gap>0:
            df1 = self.merge_called_blocks(df1)

    ########################### Writing IBD2 blocks to pandas dataframe ###############################
        starts_ibd2, ends_ibd2 = self.ibd_stat_to_block(ibd2)
        l_ibd2 = ends_ibd2 - starts_ibd2
        ends_map_ibd2 = r_map[ends_ibd2 - 1]  # -1 to stay within bounds
        starts_map_ibd2 = r_map[starts_ibd2]
        l_map_ibd2 = ends_map_ibd2 - starts_map_ibd2

        df2 = self.create_df(starts_ibd2, ends_ibd2, starts_map_ibd2, ends_map_ibd2, 
                            l_ibd2, l_map_ibd2, self.ch, self.min_cm2,
                            iid1, iid2, segment_type='IBD2')
        if self.output:
            print(f"Called n={len(df2)} IBD2 Blocks > {self.min_cm2} cM")
            l = np.max(df2["lengthM"])
            print(f"Longest Block: {l *100:.2f} cM")
        # Merge Blocks in Postprocessing Step
        if self.max_gap>0:
            df2 = self.merge_called_blocks(df2)

        # merge IBD2 blocks to the same dataframe as IBD1 blocks
        df = pd.concat((df1, df2), ignore_index=True)

        if self.save==1:
            self.save_output(df)
        elif self.save==2:
            self.save_output(df, r_map=r_map, post=post0[0,:])
        elif self.save==3:
            self.save_output(df, r_map=r_map, post=post0)
        return df, r_map, post0
    
def load_Postprocessing(p_model="hapROH"):
    """Factory Method for PostProcessing class"""
    if p_model == "hapROH":
        pp = PostProcessing()
    elif p_model == "None":
        pp = NoPostProcessing()
    elif p_model == 'IBD2':
        pp = IBD2Postprocessing()
    else:
        raise RuntimeError(f"Postprocessing method {p_model} not available!")
    return pp


# run ancIBD given an input vcf file

import argparse
from ancIBD.run import hapBLOCK_chroms
from ancIBD.IO.prepare_h5 import vcf_to_1240K_hdf
from pathlib import Path
import h5py
import os


def main():
    parser = argparse.ArgumentParser(description='Run ancIBD.')
    parser.add_argument('--vcf', action="store", dest="vcf", type=str, required=False,
                        help="path to the imputed vcf file")
    parser.add_argument('--h5', action="store", dest="h5", type=str, required=False,
                        help="path to hdf5 file. If specified, ancIBD will skip the vcf to hdf5 conversion step. \
                        Only one of --vcf and --h5 should be specified.\
                        But please make sure that the hdf5 file has suffix ch{chromosome number}.h5 (e.g, test.ch20.h5).")
    parser.add_argument('--ch', action="store", dest="ch", type=int, required=True, help='chromosome number (1-22).')
    parser.add_argument('--marker_path', action="store", dest="marker_path", type=str, required=False, help='path to the marker file')
    parser.add_argument('--map_path', action="store", dest="map_path", type=str, required=False, help='path to the map file')
    parser.add_argument('--af_path', action="store", dest="af_path", type=str, required=False, default="", help='path to the allele frequency file (optional)')
    parser.add_argument('--af_column', action='store', dest='af_column', type=str, required=False, default='', help='column name of the allele frequency in the hdf5. For example, "variants/AF_ALL" or "variants/AF_SAMPLE".')
    parser.add_argument('--out', action="store", dest="out", type=str, required=False, help='output folder to store IBD results and the intermediary .hdf5 file. If not specified, the results will be stored in the same folder as the input vcf file.')
    parser.add_argument('--prefix', action="store", dest="prefix", type=str, required=False,
                        help="prefix of output file. If not specified, the prefix will be the same as the input vcf")
    parser.add_argument('--min', action="store", dest="min", type=float, required=False, default=8, help="minimum length of IBD segment in cM. Default is 8.")
    parser.add_argument('--iid', action="store", dest="iid", type=str, required=False, help="A list of sample iids to run ancIBD on (each line contains one sample IID). The sample list must match the sample name in the provided vcf file. If unspecified, ancIBD will run on all samples in the vcf file")
    parser.add_argument('--pair', action="store", dest="pair", type=str, required=False, help="A list of sample pairs to run ancIBD on (each line contains two sample IIDs separated by a whitespace). The sample list must match the sample name in the provided vcf file, and, if --iid is specified, all samples must also appear in the iid file. If unspecified, ancIBD will run on all pairs of samples in the vcf file")
    args = parser.parse_args()

    if args.vcf is None and args.h5 is None:
        raise ValueError("One of --vcf and --h5 must be specified.")
    elif args.vcf is not None and args.h5 is not None:
        raise ValueError("Only one of --vcf and --h5 should be specified.")

    ch = args.ch

    # determine prefix of output file
    prefix = args.prefix
    if prefix is None:
        prefix = f'ch{ch}'
    else:
        prefix += f'.ch{ch}'

    # determine output folder
    oDir = args.out
    if oDir is None:
        oDir = Path(args.vcf).parent.absolute() if args.vcf is not None else Path(args.h5).parent.absolute()
    else:
        # create the output folder if it does not already exist
        if os.path.isdir(oDir) is False:
            # create the output folder and its parental folders if they do not exist
            Path(oDir).mkdir(parents=True)

    if args.vcf is not None:
        path_vcf_1240k = os.path.join(f"{oDir}", f"{prefix}.1240k.vcf")
        path_h5 = os.path.join(f"{oDir}", f"{prefix}.h5")
        col_sample_af = "" if len(args.af_path) > 0 else "AF_SAMPLE"
        if len(args.af_path) == 0:
            print("WARNING: allele frequency file not provided. The allele frequency will be calculated from the input vcf file. Make sure you have enough sample size to have good estimates of allele frequencies ")
            print('The sample allele frequency will be stored in the "variants/AF_SAMPLE" column of the output hdf5 file.')
        vcf_to_1240K_hdf(in_vcf_path = args.vcf,
                 path_vcf = path_vcf_1240k, path_h5 = path_h5,
                 marker_path = args.marker_path,
                 map_path = args.map_path,
                 af_path = args.af_path,
                 col_sample_af = col_sample_af,
                 buffer_size=20000, chunk_width=8, chunk_length=20000,
                 ch=ch)
        os.remove(path_vcf_1240k) # remove intermediary vcf files
    else:
        path_h5 = args.h5

    if args.iid is None:
        with h5py.File(path_h5, 'r') as f:
            iids = f['samples'][:].astype('str')
    else:
        iids = []
        with open(args.iid, 'r') as f:
            for line in f:
                iids.append(line.strip())
    
    run_iids = []
    if not args.pair is None:
        with open(args.pair, 'r') as f:
            for line in f:
                id1, id2 = line.strip().split()
                run_iids.append((id1, id2))
    
    # I set folder_out to an empty string because I save the output file separately in the code below
    p_col = args.af_column
    if len(p_col) == 0:
        p_col = 'variants/AF_ALL' if len(args.af_path) > 0 else 'variants/AF_SAMPLE'
    df_ibd = hapBLOCK_chroms(folder_in=path_h5[:2+path_h5.rfind('ch')],
                             iids=iids, run_iids=run_iids,
                             ch=ch, folder_out="",
                             output=False, prefix_out='', logfile=False,
                             l_model='h5', e_model='haploid_gl', h_model='FiveStateScaled', t_model='standard', p_col=p_col,
                             ibd_in=1, ibd_out=10, ibd_jump=400,
                             min_cm=args.min, cutoff_post=0.99, max_gap=0.0075)
    
    df_ibd.to_csv(os.path.join(f"{oDir}", f"{prefix}.tsv"), sep='\t', index=False)
    



"""
Functions to prepare parameters for a batch run.
Splits up all pairs into batches, in a standard algorithm
@ Author: Harald Ringbauer, 2022
"""

import numpy as np
import pandas as pd
import os as os
import itertools as it

from ancIBD.IO.ind_ibd import create_ind_ibd_df, filter_ibd_df # Post-process IBD list to individual

#########################################################

def get_iids(path_meta="/n/groups/reich/hringbauer/git/yamnaya/data/meta_v2.tsv", min_snps = 600000):
    """Return list of iids to run"""
    df1 = pd.read_csv(path_meta, sep="\t")
    df1 = df1[df1["include"]==1].copy().reset_index(drop=True) # Only include unique,unrelated samples
    if min_snps>0:
        df2 = df1[df1["n_cov_snp"]>min_snps].copy()
        print(f"Loaded {len(df2)}/{len(df1)} samples with min snps: >{min_snps}")
    else:
        df2 = df1.copy()
        print(f"Loaded {len(df2)}/{len(df1)} samples with include==1")
    
    iids = df2["iid"].values
    return iids

def get_batch_idcs(i, batch_size):
    """Return the Index of batch and the within batch index"""
    iw = i % batch_size       # Within Index
    ib = int(i / batch_size)  # Batch index
    return iw, ib

def get_batch_pair_idx(i, batch_nr):
    """Return the Index of the two batches two run,
    using only triangular comparisons"""
    l1, l2 = np.tril_indices(n=batch_nr, k=0)
    return (l1[i], l2[i])

def create_savepath(folder_base="", ch=1, b1=1, b2=2, output=True):
    """Create the savepath in standardized output format"""
    savepath = os.path.join(folder_base, f"batch{b1}_{b2}", "ch" + str(ch) + ".tsv") # 
    if output:
        print(f"Created Save Path for IBD table: {savepath}")
    return savepath

def clean_double(run_idc=[], output=True):
    """Removes self comparisons as well as double comparisons.
    Input: list of pair-wise IIDs to run
    Return cleaned list."""
    idx_include = (run_idc[:,0] < run_idc[:,1])
    run_idc = run_idc[idx_include]
    if output:
        print(f"Filtered to {np.sum(idx_include)}/{len(idx_include)} pairs.")
    return run_idc

def get_idx_batch(b, batch_size):
    """Return all the indices of Indivdiuals of batch"""
    l,u = b*batch_size,(b+1)*batch_size
    return l,u

def get_unique_iid_pairs(iids, b1, b2, batch_size):
    """Get List of unique iid pairs to run.
    Return list of unique pairs, and list of all iids"""
    if b1>b2:
        l1,u1 = get_idx_batch(b1, batch_size)
        l2,u2 = get_idx_batch(b2, batch_size)
        iids1, iids2 = iids[l1:u1], iids[l2:u2]
        iids = np.concatenate((iids1, iids2)) # Append them to each other
        run_iids =  [[i1, i2] for i1 in iids1 for i2 in iids2] # All possible combinations. 
        
    elif b1==b2:
        l,u = get_idx_batch(b1, batch_size)
        iids = iids[l:u]
        run_iids = list(it.combinations(iids, 2))
    else:
        raise RuntimeError(f"Serious error, batch-index {b1}-{b2} not valid.")
    return iids, run_iids

def get_run_lists_batch(i = 67, k=3500, batch_size = 400, output=True):
    """Get IID list and Run lists for batches of samples.
    i: Run number
    k: number of total indivdiuals
    batch_size: number of individuals in one batch
    Return batch indices """
    batch_nr = int(k/batch_size)+1 # The number of batches to run
    pw_batch_nr = int(batch_nr * (batch_nr+1)/2)

    iw, ch = get_batch_idcs(i, pw_batch_nr) # Within batch number
    ch = ch+1 # For Autosome
    b1, b2 = get_batch_pair_idx(iw, batch_nr)

    if output:
        print(f"Run Nr: {i}")
        print(f"# of batches of individuals: {batch_nr}")
        print(f"# of pw. batches to run per chromosome: {pw_batch_nr}")
        print(f"Running Chromosome: {ch}")
        print(f"Running Batch Pair: {b1}-{b2}")
    return b1, b2, ch

def save_ibd_df(df_ibd, savepath, create=True):
    """Saves IBD Dataframe"""
    if create:
        folder = os.path.dirname(savepath)
        if not os.path.exists(folder):
            os.makedirs(folder)
    df_ibd.to_csv(savepath, sep="\t", index=False) # Save the Values
    print(f"Saved {len(df_ibd)} IBD blocks.")
    
def get_run_params_from_i(i, metapath="./data/iid_lists/iid_ibd_eurasia_v1.tsv", batch_size = 400, min_snps=0, 
                         output = True, folder_out = "/n/groups/reich/hringbauer/git/ibd_euro/output/ibd/v1/"):
    """Return the run parameters for run i
    min_snps: Minimum number of SNPs covered (for potential filtering on meta file)
    Returns iids, run_iids, and the output folder"""
    iids = get_iids(path_meta=metapath, min_snps=min_snps)
    b1, b2, ch = get_run_lists_batch(i = i, k=len(iids), batch_size = batch_size, output=output)
    iids, run_iids = get_unique_iid_pairs(iids, b1, b2, batch_size)
    path_ibd = create_savepath(folder_out, ch=ch, b1=b1, b2=b2, output=output)
    return iids, run_iids, ch, path_ibd

######################################################################################################
######################################################################################################
### Post-process batch runs

def join_chromosomes(base_path, chs=range(1,23), file_out="ch_all.tsv", output=True):
    """Join different Chromosomes together and save output.
    Return joined dataframe.
    file_out: If given, save the joint file with that name into the base_path folder"""
    df_ibds = []
    for ch in chs:
        path_save = os.path.join(base_path, f"ch{ch}.tsv")
        if not os.path.exists(path_save):
            print(f"Warning: {path_save} does not exist. Skipping!")
        else:
            df = pd.read_csv(path_save, sep="\t")
            if output:
                print(f"Chromosome {ch}; Loaded {len(df)} IBD")
            df_ibds.append(df)
    df_ibds = pd.concat(df_ibds) # Join together
    
    if len(file_out)>0:
        path_save = os.path.join(base_path, file_out)
        df_ibds.to_csv(path_save, sep="\t", index=False) # Save the Values
        if output:
            print(f"Saved {len(df_ibds)} IBD to {path_save}.")
    return df_ibds

def to_ind_df_batch(b1, b2, folder_out = "/n/groups/reich/hringbauer/git/ibd_euro/output/ibd/v1/",
                    chs=range(1,23), min_cms = [8, 12, 16, 20], snp_cm = 220, min_cm = 8, output=False):
    """Post-process a batch of individals.
    Returns individal IBD dataframe"""

    folder_batch = os.path.join(folder_out, f"batch{b1}_{b2}/")

    df_ibds = join_chromosomes(folder_batch, file_out="", chs=chs, output=output)

    df_res = create_ind_ibd_df(ibd_data=df_ibds,
                               min_cms = min_cms, snp_cm = snp_cm, min_cm = min_cm, sort_col = 0, 
                               output=output, savepath = "")
    return df_res

def to_ind_df_batches(batches=8, folder_out = "/n/groups/reich/hringbauer/git/ibd_euro/output/ibd/v1/",
                      chs=range(1,23), min_cms = [8, 12, 16, 20], snp_cm = 220, min_cm = 8, output=False,
                      savepath=""):
    """Runs multiple combinations of batches (wrapper of single batch function and then combines the processed batches. 
    Postprocess IBD to individal summary dataframe.
    Return merged IBD dataframe
    batches: If int: create all possible combinations. Otherwise needs to be array [n,2] of all
    pairs to run.
    savepath: If given, save IBD dataframe to there"""
    if isinstance(batches, int):
        batches = np.column_stack(np.tril_indices(n=batches, k=0))
    res = []
    for b1,b2 in batches:
        df_res = to_ind_df_batch(b1, b2, folder_out = folder_out, chs=chs,
                                 min_cms = min_cms, snp_cm = snp_cm, 
                                 min_cm = min_cm, output=output)
        res.append(df_res)
    df_res = pd.concat(res)
    return df_res 

def to_ibd_df_batches(batches=8, folder_out = "/n/groups/reich/hringbauer/git/ibd_euro/output/ibd/v1/",
                      chs=range(1,23), min_cms = [8, 12, 16, 20], snp_cm = 220, min_cm = 8, output=False,
                      savepath=""):
    """Runs multiple combinations of batches (wrapper of single batch function and then combines the processed batches. 
    Postprocess IBD to individal summary dataframe.
    Return merged IBD dataframe
    batches: If int: create all possible combinations. Otherwise needs to be array [n,2] of all
    pairs to run.
    savepath: If given, save IBD dataframe to there"""
    if isinstance(batches, int):
        batches = np.column_stack(np.tril_indices(n=batches, k=0))
    res = []
    
    for b1,b2 in batches:
        folder_batch = os.path.join(folder_out, f"batch{b1}_{b2}/")
        df_ibds = join_chromosomes(folder_batch, file_out="", 
                                   chs=chs, output=output)
        res.append(df_ibds)
    df_res = pd.concat(res)
    df_res = filter_ibd_df(df_res, min_cm=min_cm, snp_cm=snp_cm, output=output) 
    return df_res 


def print_runid_missing(b = 1, folder_out = "", output=False):
    """Finds and prints indices of missing output (chXX.tsv) for batchwise runs.
    Return list of missing indices. Ideal for rerunning batch scripts.
    Uses C Indexing as would be used in submission script."""
    batches = np.column_stack(np.tril_indices(n=b, k=0))

    ls = []
    for ch in range(1,23):
        for b1,b2 in batches:
            folder_batch = os.path.join(folder_out, f"batch{b1}_{b2}/ch{ch}.tsv")
            exist = os.path.exists(folder_batch)
            if not exist:
                if output:
                    print(f"Missing! ch: {ch} batch: {b1}-{b2}")
                l = b * (b+1)/2 * (ch-1) + (b1+1) * b1/2 + b2 + 1 # The +1 is c indexing
                ls.append(str(int(l)))
    return ls



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import socket as socket
import os as os
import sys as sys
import multiprocessing as mp
import h5py
import allel

#sys.path.append("/n/groups/reich/hringbauer/git/hapBLOCK/python3/")  # hack to get development package first in path
from ancIBD.IO.h5_load import get_coverage,get_genos_pairs,get_idx_iid,get_idx_iid_exact, get_markers_good, get_opp_homos_f

def opp_homos(g1, g2):
    """Return opposing homozygotes"""
    o1 = (g1 == 0) & (g2 == 2)
    o2 = (g1 == 2) & (g2 == 0)
    return (o1 | o2)

def plot_hets(map_het=[], het=[], ax=None, het_c="slateblue", c_roh="seagreen", 
              figsize=(14,2), cm_lim=[], ylim=[-0.1, 1.1], fs = 12,
              alpha=0.3, ms=1, lw = 12, title="", min_cm=0, xtickl=True,
              xlabel="centimorgan", ylabel = f"Opp. Homozygote (y/n)", 
              plot=True, savepath=""):
    """Plot Heterozygotes against genenetic map,
    plus ROH calls.
    lw: Linewidth of ROH""" 
    if ax==None:
        plt.figure(figsize=figsize)
        ax = plt.gca()
    ax.plot(map_het*100, (het * 1.1 - 0.05), "o", ms=ms, alpha=alpha, zorder=0, color=het_c)
    ax.set_xlabel(xlabel, fontsize=fs)
    if not xtickl:
        ax.set_xticklabels([])
    ax.set_ylim(ylim)
    
    ax2 = ax.twinx()
    ax2.set_ylim(ax.get_ylim())
    ax2.set_yticks(np.array([1,0]) * 1.1 - 0.05)
    ax2.set_yticklabels([])
    ax.set_yticklabels([])
    ax2.set_ylabel(ylabel, fontsize=fs*0.8, color=het_c, rotation=270, labelpad=fs)
    
    if len(cm_lim)==2:
        ax.set_xlim(cm_lim)
        
    if len(title)>0:
        ax.set_title(title, fontsize=fs)
        
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches = 'tight', pad_inches = 0, dpi=300)
        print(f"Saved figure to: {savepath}")
        #plt.savefig(folder + "posterior_cm.png", bbox_inches = 'tight', pad_inches = 0, dpi=300)
    
    if plot:
        plt.show()  
        
def plot_opp_homos(path_h5 = f"/n/groups/reich/hringbauer/hapsburg_runs/data/data_eirini/h5/all_ch", ch = 16,
                  iids = ["MYG001.A0101", "MYG006.A0101"], cutoff=0.99, output=True, exact=True,
                  cm_lim=[], ms=4, savepath=""):
    """Plot pairwise opp. homoyzgotes. Loads f5 data, and calls and plots opp. homos"""
    path_h5 = f"{path_h5}{ch}.h5"
    f = h5py.File(path_h5, "r") # Load for Sanity Check. See below!
    g1, g2, m = get_genos_pairs(f, sample1=iids[0], sample2=iids[1], 
                                cutoff=cutoff, output=output, exact=exact)
    o_homos = opp_homos(g1, g2)
    plot_hets(m, o_homos, title=f"{iids[0]} - {iids[1]}, Chromosome {ch}, Opposing Homozygotes", ms=ms,
              cm_lim=cm_lim, savepath=savepath) # ./figs/principle_proof/siblings.png
    
def plot_identical_gts(path_h5 = f"/n/groups/reich/hringbauer/hapsburg_runs/data/data_eirini/h5/all_ch", ch = 16,
                  iids = ["MYG001.A0101", "MYG006.A0101"], cutoff=0.99, output=True, exact=True,
                  cm_lim=[], ms=4, savepath=""):
    """Plot pairwise opp. homoyzgotes. Loads f5 data, and calls and plots opp. homos"""
    path_h5 = f"{path_h5}{ch}.h5"
    f = h5py.File(path_h5, "r") # Load for Sanity Check. See below!
    g1, g2, m = get_genos_pairs(f, sample1=iids[0], sample2=iids[1], 
                                cutoff=cutoff, output=output, exact=exact)
    diff = (g1!=g2)
    plot_hets(m, diff, title=f"{iids[0]} - {iids[1]}, Chromosome {ch}, Different Diploid Genotyeps", ms=ms,
              cm_lim=cm_lim, ylabel = f"Diff. Genotypes (y/n)", savepath=savepath) # ./figs/principle_proof/siblings.png


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import socket as socket
import os as os
import sys as sys
import h5py
import matplotlib.cm as cm
import itertools as it
import multiprocessing as mp
import matplotlib.patches as mpatches
from scipy.special import gammaincinv
from matplotlib import gridspec
import matplotlib.patheffects as pe
from matplotlib import rcParams
rcParams['font.family'] = 'sans-serif'   # Set the default
rcParams['font.sans-serif'] = ['Arial']  # Make sure to have the font installed (it is on cluster for Harald)

#sys.path.insert(0,"/n/groups/reich/hringbauer/git/hapBLOCK/python3/")  # hack to get development package first in path
from ancIBD.ibd_stats.funcs import give_sub_df

###########################################################################
###########################################################################
### The Functions for the Plot

def chrom_length(ch, ch_lengths=[], output=False):
    """Get and return length of Chromosome ch
    If ch_lengths not given, use default (from Eigenstrat map).
    Atm only do autosomes!"""
    if len(ch_lengths)==0:
        ch_lengths = [2.86273, 2.688325, 2.232573, 2.145423,
                      2.040858, 1.920325, 1.871526, 1.680022,
                      1.66301, 1.809153, 1.582171, 1.746799,
                      1.257046, 1.202023, 1.41346, 1.340263,
                      1.284738, 1.177099, 1.077316, 1.082134,
                      0.627865, 0.740762]
        return ch_lengths[ch-1]

def load_bad_areas(path="./Data/1000Genomes/Markers/1240k/snp_density.csv", min_snps=50):
    """Load areas of low SNP density, and return list of Dataframes 
    (one for each chromosome)"""
    df_lows = []
    
    for i in range(1,23):
        df_t = pd.read_csv(path, sep="\t")
        df_t = df_t[df_t["chr"]==i]
        df_t = df_t[df_t["counts"]<min_snps]
        df_lows.append(df_t)
    return df_lows

def plot_chromosome(ax, l, x_pos, lw=24, df_roh = [], df_low = [], 
                    c_back="lightgray", c_ibd="maroon"):
    """Plot a Chromosome of length l with centromer ctr on ax 
    at x_pos"""
    ln, = ax.plot([x_pos, x_pos], [-0.05,l+0.05], lw = lw, color=c_back,
                      solid_capstyle = 'round', zorder=0,
                  path_effects=[pe.Stroke(linewidth=lw+3, foreground='k'), pe.Normal()])
    
    ### Plot the ROH List if given
    if len(df_roh) > 0:
        starts, ends = df_roh["StartM"].values, df_roh["EndM"].values
        
        for i in range(len(df_roh)):
            ax.plot([x_pos, x_pos], [starts[i], ends[i]], lw=lw, color=c_ibd, 
                    zorder=1, alpha=1.0, solid_capstyle="butt")
    
    ### Plot shadows of bad areas
    if len(df_low)>0:
        starts, ends = df_low["StartM"].values, df_low["EndM"].values
        
        for i in range(len(df_low)):
            ax.plot([x_pos, x_pos], [starts[i], ends[i]], lw=lw, color="k", 
                    zorder=2, alpha=0.8, solid_capstyle="butt")
        
            
def plot_segments_individual(df_ibd, fs=12, figsize=(8,8), savepath="", 
                            plot_bad=True, title="", plot=True, 
                            c_back="lightgray", c_ibd="maroon"):
    """Plot ROH in one ancient Individual.
    gap: What Gap to Merge [in cM!]
    prefix_out: If there is a folder before indivual data, e.g. e01/
    c_back: Background Karyotype color.
    c_ibd: Foreground IBD"""
    
    ### Load the Data (could be done seperately)
    df_lows=[[] for _ in range(23)] ### Plot the excluded areas. Placeholder
    
    plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 2])
    gs.update(hspace=0.1) # set the spacing between axes. 

    ax = plt.subplot(gs[0]) # The left subplot
    ax1 = plt.subplot(gs[1]) # The middle one
        
    ### Plot the First 11 Chromosomes
    for i in range(1,12):
        l = chrom_length(i)
        plot_chromosome(ax, l, x_pos=(i), df_roh=df_ibd[i-1], df_low=[],
                        c_back=c_back, c_ibd=c_ibd)  
    
    ### Plot the Second 11 Chromosomes
    for i in range(12,23):
        l = chrom_length(i)
        plot_chromosome(ax1, l, x_pos=(i - 11), df_roh=df_ibd[i-1], df_low=[],
                        c_back=c_back, c_ibd=c_ibd)

    ### Set the Plot x and y Limits
    ax.set_xlim([0.3, 11.5])
    ax1.set_xlim([0.3, 11.5])

    ax.set_ylim([-0.3,3.3])
    ax1.set_ylim([-0.3, 2.05])

    ### Turn off the Y Axis
    for ax0 in [ax, ax1]:
        ax0.spines['right'].set_visible(False)
        ax0.spines['top'].set_visible(False)
        ax0.spines['bottom'].set_visible(False)
        ax0.yaxis.set_tick_params(labelsize=fs)

    rang = np.arange(1,12)
    ax.set_xticks(rang)
    ax.set_xticklabels(rang, fontsize=fs)

    ax1.set_xticks(rang)
    ax1.set_xticklabels(rang+11, fontsize=fs)
    ax1.set_xticklabels(rang+11, fontsize=fs)
    
    if len(title)>0:
        ax.set_title(title, fontsize=fs*2)
        
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches = 'tight', pad_inches = 0, dpi=300)
        print(f"Saved figure to {savepath}")
    if plot:
        plt.show()
        
def plot_karyo_from_ibd_df(df_ibd, iids=[], min_cm=12,
                           c_back="silver", c_ibd="darkslateblue",
                           savepath="", fs=13, figsize=(7,7), 
                           title="", **kwargs):
    """Make Karyotype Plot from IBD dataframe.
    df_ibd: List of IBD or ROH in standard hapROH, bapBLOCK format
    iids: [iid1, iid2]
    min_cm: Minimum Length of IBDs to plot [in cM]
    savepath: Where to save individuals
    title: What title to add.
    kwargs: Pass on Keyword arguments to the plotting function"""
    iid1, iid2 = iids[0], iids[1]
    df_target = give_sub_df(df_ibd, pop1=iid1, pop2=iid2, col="iid")
    df_target = df_target[df_target["lengthM"]>min_cm/100.0]
    df_ibds = [df_target[df_target["ch"]==i] for i in range(1,23)] # Split up per Chromosome
    
    plot_segments_individual(df_ibds, fs=fs, figsize=figsize,
                             c_back=c_back, c_ibd=c_ibd, savepath=savepath, 
                             title=title, **kwargs)
    
##########################################################
### Plot Length Distributions

def expected_block_pdf(x, chr_l, m):
    """Gives back the pdfs for Blocks of Length l [Morgan]
    on a Chromosome of Length [Morgan].
    m: Nr of Meiosis.
    Return PDF (per Morgan)"""
    pdf0 = (chr_l - x) * m**2 * np.exp(-x * m)
    pdf1 = 2 * np.exp(- x * m) * m 
    return (pdf0 + pdf1) * (x < chr_l)  # If x < chr_l return 0

def expected_block_pdf_chromosomes(x, chr_lgts, m):
    """Calculate the PDF of ROH blocks of length x [Morgan]
    for m Recombination events
    x: Can be Array
    chr_lgts: Array of all chromosome lengths [in Morgan]
    Return PDF (per Morgan)"""
    pdfs = [expected_block_pdf(x, chr_l, m) for chr_l in chr_lgts]
    pdf_full = np.sum(pdfs, axis=0)
    return pdf_full

def coal_prob(m, comm_anc=1):
    """Calculate Coalescence Probability.
    m: Nr of Meiosis
    comm_anc: How many common ancestors"""
    c_prob = comm_anc * (1 / 2) ** m
    return c_prob

def exp_blocks_full_individual(x, m, comm_anc=1):
    """Calculates the Expected Nr of ROH Blocks per Morgan for full Individual
    x: Array of Block Lenths [in Morgan]
    m: Nr of Meisois
    comm_anc: Nr of Ancestry Loops"""
    
    chr_lgts = [chrom_length(ch) for ch in range(1,23)]
    
    pdf_full = expected_block_pdf_chromosomes(x, chr_lgts, m)
    c_prob = coal_prob(m, comm_anc)
    exp_blocks = pdf_full * c_prob
    return exp_blocks


def plot_pde_individual(df_ibd, figsize=(8,6), bw_cm=4,
                        plotlim=[4,100], savepath="", 
                        output=False, lw_curve=3,
                        comm_ancs=[4,4,4,2], ms=[4,6,5,4], 
                        labels=["First Cousins", "Second Cousins", "5 generations anc.", "4 generations anc."],
                        cs=["red", "green", "orange", "gray"], ls=[], title="", leg_loc="upper right"):
    """Plot Histograms/PDEs of IBD Distribution for one Individual (iid)
    bw_cm: Length of one Bin (in cM)
    comm_ancs: How many common ancestors to plot [list]
    ms: How many meiosis to plot [list]
    labels: what labels do they have [list]
    cs: what colors to plot [list]"""
    if len(ls)==0:
        ls = ["-" for _ in cs]
    ### Load And Prepare Data
    bins = np.arange(plotlim[0], plotlim[1], bw_cm)
    bin_mean = (bins[1:] + bins[:-1]) / 2.0  # Mean of each bin
    
    ### Do the Plot
    fs = 16

    plt.figure(figsize=figsize)
    ax=plt.gca()
    ax.hist(df_ibd["lengthM"]*100, bins=bins, ec="k", fc="dodgerblue", label="Observed IBD")
    
    # Plot the Empirical Averages
    for i in range(len(labels)):
        block_pdf = exp_blocks_full_individual(bin_mean/100, m=ms[i], comm_anc=comm_ancs[i])
        ax.plot(bin_mean, bw_cm * block_pdf/100, c=cs[i], linestyle=ls[i], 
                label=labels[i], lw=lw_curve) # Plot Density Per cM (adjusted for bin width)

    plt.xlabel("IBD Length [cm]", fontsize=fs)
    plt.ylabel(f"Number per {bw_cm} cM Bin", fontsize=fs)
    plt.title(title, fontsize=fs)
    leg = plt.legend(loc = leg_loc, fontsize=fs)
    leg.set_title("Relationship:", prop = {'size':fs})
    
    plt.tick_params(axis='both', which='major', labelsize=fs)
    
    if len(savepath)>0:
        plt.savefig(savepath, bbox_inches = 'tight', pad_inches = 0, dpi=300)
        print(f"Saved to {savepath}.")
    plt.show()
    
def plot_pde_individual_from_ibd_df(df_ibd, iids=[], min_cm=12, figsize=(8,6),
                                    comm_ancs =[4,4,2,2], ms=[4,6,5,4],
                                    labels=["First Cousins", "Second Cousins", "5 generations anc.", "4 generations anc."],
                                    cs = ["red", "green", "orange", "gray"], ls = [], title="",
                                    bw_cm=4, plotlim=[4,100], lw_curve=3,
                                    leg_loc = "upper right", output=False, savepath=""):
    """df_ibd: List of IBD or ROH in standard hapROH, bapBLOCK format
       iids: [iid1, iid2]
       min_cm: Minimum Length of IBDs to plot [in M]
       savepath: Where to save individuals"""
    iid1, iid2 = iids[0], iids[1]    
    df_target = give_sub_df(df_ibd, pop1=iid1, pop2=iid2, col="iid")
    df_target = df_target[df_target["lengthM"]>min_cm/100.0]
    
    ### Do the Plot
    plot_pde_individual(df_target, figsize=figsize, bw_cm=bw_cm,
                    plotlim=plotlim,
                    output=output, lw_curve=lw_curve, 
                    comm_ancs=comm_ancs, ms=ms, labels=labels,
                    cs=cs, ls=ls, title=title, 
                    leg_loc=leg_loc, savepath=savepath)




